{- Generated by DrIFT (Automatic class derivations for Haskell) -}
{-# LINE 1 "src/C/Prims.hs" #-}
module C.Prims where

import Data.Binary
import Data.Monoid
import Data.Typeable

import Doc.DocLike
import Doc.PPrint
import PackedString
import StringTable.Atom
import qualified Cmm.Op as Op

data PrimTypeType = PrimTypeIntegral | PrimTypeFloating | PrimTypePointer | PrimTypeVoid
    deriving(Show,Eq,Ord)

data PrimType = PrimType {
    primTypeName :: ExtType,
    primTypeType :: {-# UNPACK #-} !PrimTypeType,
    primTypeAlignmentOf :: {-# UNPACK #-} !Int,
    primTypeIsSigned :: {-# UNPACK #-} !Bool,
    primTypeSizeOf :: {-# UNPACK #-} !Int
    } deriving(Show)

type ExtType = String

instance Show Requires where
    show (Requires [] []) = "()"
    show (Requires xs ys) = show (xs,ys)

data Requires = Requires {
    reqIncludes :: [String],
    reqLibraries :: [String]
    } deriving(Eq, Ord)
    {-! derive: Monoid, Binary !-}

data DotNetPrim = DotNetField | DotNetCtor | DotNetMethod
    deriving(Typeable, Eq, Ord, Show)
    {-! derive: Binary !-}

data Prim =
    PrimPrim Atom          -- Special primitive implemented in the compiler somehow.
    | CConst { primConst :: !PackedString }  -- C code which evaluates to a constant
    | Func {
        funcIOLike :: {-# UNPACK #-} !Bool,
        funcName :: !PackedString,
        primArgTypes :: [ExtType],
        primRetType :: ExtType
        }   -- function call with C calling convention
    | IFunc {
        funcIOLike :: {-# UNPACK #-} !Bool,
        primArgTypes :: [ExtType],
        primRetType :: ExtType
        } -- indirect function call with C calling convention
    | AddrOf !PackedString          -- address of linker name
    | Peek { primArgTy :: Op.Ty }  -- read value from memory
    | Poke { primArgTy :: Op.Ty }  -- write value to memory
    | PrimTypeInfo {
        primArgTy :: Op.Ty,
        primRetTy :: Op.Ty,
        primTypeInfo :: {-# UNPACK #-} !PrimTypeInfo
        }
    | PrimString !PackedString                                 -- address of a raw string. encoded in utf8.
    | PrimDotNet {
        primStatic :: {-# UNPACK #-} !Bool,
        primDotNet :: !DotNetPrim,
        primIOLike :: {-# UNPACK #-} !Bool,
        primAssembly :: !PackedString,
        primDotNetName :: !PackedString
        }
    | Op {
        primCOp :: Op.Op Op.Ty,
        primRetTy :: Op.Ty
        }
    deriving(Typeable, Eq, Ord, Show)
    {-! derive: Binary !-}

data PrimTypeInfo = PrimSizeOf | PrimMaxBound | PrimMinBound | PrimAlignmentOf | PrimUMaxBound
    deriving(Typeable, Eq, Ord, Show)
    {-! derive: Binary !-}

primStaticTypeInfo :: Op.Ty -> PrimTypeInfo -> Maybe Integer
primStaticTypeInfo (Op.TyBits (Op.Bits b) _) w = Just ans where
    bits = toInteger b
    ans = case w of
        PrimSizeOf -> bits `div` 8
        PrimAlignmentOf ->  bits `div` 8
        PrimMinBound -> negate $ 2^(bits - 1)
        PrimMaxBound -> 2^(bits - 1) - 1
        PrimUMaxBound -> 2^bits - 1
primStaticTypeInfo _ _ = Nothing

-- | These primitives may safely be duplicated without affecting performance or
-- correctness too adversly. either because they are cheap to begin with, or
-- will be recombined in a later pass.

primIsCheap :: Prim -> Bool
primIsCheap AddrOf {} = True
primIsCheap CConst {} = True
primIsCheap PrimString {} = True
primIsCheap PrimTypeInfo {} = True
primIsCheap Op { primCOp = op } = Op.isCheap op
primIsCheap _ = False

aprimIsCheap (APrim p _) = primIsCheap p

-- | whether a primitive represents a constant expression (assuming all its arguments are constant)
-- TODO needs grin support
primIsConstant :: Prim -> Bool
primIsConstant CConst {} = True
primIsConstant AddrOf {} = True
primIsConstant PrimString {} = True
primIsConstant PrimTypeInfo {} = True
primIsConstant Op { primCOp = op } = Op.isEagerSafe op
primIsConstant _ = False

-- | whether a primitive can be eagarly evaluated.
-- TODO needs grin support
primEagerSafe :: Prim -> Bool
primEagerSafe CConst {} = True
primEagerSafe PrimString {} = True
primEagerSafe AddrOf {} = True
primEagerSafe PrimTypeInfo {} = True
primEagerSafe Op { primCOp = op } = Op.isEagerSafe op
primEagerSafe _ = False

parsePrimString s = do
    ws@(_:_) <- return $ words s
    let v = case last ws of
            '&':s -> AddrOf (packString s)
            s -> Func False (packString s) [] ""
    let f opt@('-':'l':_) = Requires [] [opt]
        f s = Requires [s] []
    return (APrim v (mconcat (map f (init ws))))

primPrim s = APrim (PrimPrim $ toAtom s) mempty

data APrim = APrim Prim Requires
    deriving(Typeable,  Eq, Ord)
    {-! derive: Binary !-}

instance Show APrim where
    showsPrec n (APrim p r) | r == mempty = showsPrec n p
    showsPrec n (APrim p r) = showsPrec n p . shows r

instance PPrint d Prim  => PPrint d APrim where
    pprintAssoc a n (APrim p _) = pprintAssoc a n p

instance DocLike d => PPrint d Prim where
    pprint (PrimPrim t) = text (fromAtom t)
    pprint (CConst s) = parens (text $ unpackPS s)
    pprint (Func _ s xs r) = parens (text r) <> text (unpackPS s) <> tupled (map text xs)
    pprint (IFunc _ xs r) = parens (text r) <> parens (char '*') <> tupled (map text xs)
    pprint (AddrOf s) = char '&' <> text (unpackPS s)
    pprint (PrimString s) = tshow s <> char '#'
    pprint (Peek t) = char '*' <> tshow t
    pprint (Poke t) = char '=' <> tshow t
    pprint Op { primCOp = Op.BinOp bo ta tb, primRetTy = rt } | rt == ta && rt == tb = parens (pprint rt) <> tshow bo
    pprint Op { primCOp = Op.UnOp bo ta, primRetTy = rt } | rt == ta = parens (pprint rt) <> tshow bo
    pprint Op { primCOp = op, primRetTy = rt } = parens (pprint rt) <> pprint op
    pprint PrimDotNet { primDotNet = dn,  primDotNetName = nn} = parens (text (unpackPS nn))
    pprint PrimTypeInfo { primArgTy = at, primTypeInfo = PrimSizeOf } = text "sizeof" <> parens (tshow at)
    pprint PrimTypeInfo { primArgTy = at, primTypeInfo = PrimAlignmentOf } = text "alignmentof" <> parens (tshow at)
    pprint PrimTypeInfo { primArgTy = at, primTypeInfo = PrimMaxBound } = text "max" <> parens (tshow at)
    pprint PrimTypeInfo { primArgTy = at, primTypeInfo = PrimUMaxBound } = text "umax" <> parens (tshow at)
    pprint PrimTypeInfo { primArgTy = at, primTypeInfo = PrimMinBound } = text "min" <> parens (tshow at)

instance DocLike d => PPrint d Op.Ty where
    pprintAssoc _ n p = text (showsPrec n p "")
instance (DocLike d,Show v) => PPrint d (Op.Op v) where
    pprintAssoc _ n p = text (showsPrec n p "")

parseDotNetFFI :: Monad m => String -> m Prim
parseDotNetFFI s = ans where
    init = PrimDotNet { primIOLike = False, primStatic = False, primDotNet = DotNetField, primAssembly = packString "", primDotNetName = packString "" }
    ans = case words s of
        ("static":rs) -> f rs init { primStatic = True }
        rs -> f rs init
    f ("field":rs) dn = g dn { primDotNet = DotNetField } rs
    f ("ctor":rs) dn = g dn { primDotNet = DotNetCtor } rs
    f ("method":rs) dn = g dn { primDotNet = DotNetMethod } rs
    f _ _ = fail "invalid .NET ffi specification"
    g dn ['[':rs] | (as,']':nm) <- span (/= ']') rs = return dn { primAssembly = packString as, primDotNetName = packString nm }
    g dn [n] = return dn { primDotNetName = packString n }
    g _ _ = fail "invalid .NET ffi specification"
{-* Generated by DrIFT : Look, but Don't Touch. *-}
instance Monoid Requires where
    mempty = Requires mempty mempty
    mappend (Requires aa ab) (Requires aa' ab') = Requires (mappend aa aa')(mappend ab ab')

instance Data.Binary.Binary Requires where
    put (Requires aa ab) = do
	    Data.Binary.put aa
	    Data.Binary.put ab
    get = do
    aa <- get
    ab <- get
    return (Requires aa ab)

instance Data.Binary.Binary DotNetPrim where
    put DotNetField = do
	    Data.Binary.putWord8 0
    put DotNetCtor = do
	    Data.Binary.putWord8 1
    put DotNetMethod = do
	    Data.Binary.putWord8 2
    get = do
	    h <- Data.Binary.getWord8
	    case h of
	      0 -> do
		    return DotNetField
	      1 -> do
		    return DotNetCtor
	      2 -> do
		    return DotNetMethod
	      _ -> fail "invalid binary data found"

instance Data.Binary.Binary Prim where
    put (PrimPrim aa) = do
	    Data.Binary.putWord8 0
	    Data.Binary.put aa
    put (CConst ab) = do
	    Data.Binary.putWord8 1
	    Data.Binary.put ab
    put (Func ac ad ae af) = do
	    Data.Binary.putWord8 2
	    Data.Binary.put ac
	    Data.Binary.put ad
	    Data.Binary.put ae
	    Data.Binary.put af
    put (IFunc ag ah ai) = do
	    Data.Binary.putWord8 3
	    Data.Binary.put ag
	    Data.Binary.put ah
	    Data.Binary.put ai
    put (AddrOf aj) = do
	    Data.Binary.putWord8 4
	    Data.Binary.put aj
    put (Peek ak) = do
	    Data.Binary.putWord8 5
	    Data.Binary.put ak
    put (Poke al) = do
	    Data.Binary.putWord8 6
	    Data.Binary.put al
    put (PrimTypeInfo am an ao) = do
	    Data.Binary.putWord8 7
	    Data.Binary.put am
	    Data.Binary.put an
	    Data.Binary.put ao
    put (PrimString ap) = do
	    Data.Binary.putWord8 8
	    Data.Binary.put ap
    put (PrimDotNet aq ar as at au) = do
	    Data.Binary.putWord8 9
	    Data.Binary.put aq
	    Data.Binary.put ar
	    Data.Binary.put as
	    Data.Binary.put at
	    Data.Binary.put au
    put (Op av aw) = do
	    Data.Binary.putWord8 10
	    Data.Binary.put av
	    Data.Binary.put aw
    get = do
	    h <- Data.Binary.getWord8
	    case h of
	      0 -> do
		    aa <- Data.Binary.get
		    return (PrimPrim aa)
	      1 -> do
		    ab <- Data.Binary.get
		    return (CConst ab)
	      2 -> do
		    ac <- Data.Binary.get
		    ad <- Data.Binary.get
		    ae <- Data.Binary.get
		    af <- Data.Binary.get
		    return (Func ac ad ae af)
	      3 -> do
		    ag <- Data.Binary.get
		    ah <- Data.Binary.get
		    ai <- Data.Binary.get
		    return (IFunc ag ah ai)
	      4 -> do
		    aj <- Data.Binary.get
		    return (AddrOf aj)
	      5 -> do
		    ak <- Data.Binary.get
		    return (Peek ak)
	      6 -> do
		    al <- Data.Binary.get
		    return (Poke al)
	      7 -> do
		    am <- Data.Binary.get
		    an <- Data.Binary.get
		    ao <- Data.Binary.get
		    return (PrimTypeInfo am an ao)
	      8 -> do
		    ap <- Data.Binary.get
		    return (PrimString ap)
	      9 -> do
		    aq <- Data.Binary.get
		    ar <- Data.Binary.get
		    as <- Data.Binary.get
		    at <- Data.Binary.get
		    au <- Data.Binary.get
		    return (PrimDotNet aq ar as at au)
	      10 -> do
		    av <- Data.Binary.get
		    aw <- Data.Binary.get
		    return (Op av aw)
	      _ -> fail "invalid binary data found"

instance Data.Binary.Binary PrimTypeInfo where
    put PrimSizeOf = do
	    Data.Binary.putWord8 0
    put PrimMaxBound = do
	    Data.Binary.putWord8 1
    put PrimMinBound = do
	    Data.Binary.putWord8 2
    put PrimAlignmentOf = do
	    Data.Binary.putWord8 3
    put PrimUMaxBound = do
	    Data.Binary.putWord8 4
    get = do
	    h <- Data.Binary.getWord8
	    case h of
	      0 -> do
		    return PrimSizeOf
	      1 -> do
		    return PrimMaxBound
	      2 -> do
		    return PrimMinBound
	      3 -> do
		    return PrimAlignmentOf
	      4 -> do
		    return PrimUMaxBound
	      _ -> fail "invalid binary data found"

instance Data.Binary.Binary APrim where
    put (APrim aa ab) = do
	    Data.Binary.put aa
	    Data.Binary.put ab
    get = do
    aa <- get
    ab <- get
    return (APrim aa ab)

--  Imported from other files :-
